/* ------------------------------------------------------------- */
/* T5.cc                                                         */
/* Stefan Bauberger (bauberger@physik.uni-wuerzburg.de)          */
/* Ayres Freitas (afreitas@pitt.edu)			         */
/* last revision 27.07.18 (4.12.95)                              */
/* ------------------------------------------------------------- */
/* Scalar integral T12345 for two-loop self-energies             */
/* as part of the package s2lse for the calculation of scalar    */
/* two-loop self-energy integrals                                */
/* ------------------------------------------------------------- */
/* Limitations:                                                  */
/*    The maximal precision is about 8 digits for T12345         */
/*    and 16 digits with quad-precision,                         */
/*    depending very much on the parameter region                */
/* ------------------------------------------------------------- */
/* References:                                                   */
/*   [1] = Bauberger, Berends, Boehm, Buza,                      */
/*         hep-ph/9409388, Nucl.Phys.B 343 (1995) 383            */
/*   [2] = Bauberger, Boehm,                                     */
/*         hep-ph/9501201, Nucl.Phys.B 445 (1995) 25             */
/*   [3] = Berends, Tausk, Nucl. Phys. B 421 (1994) 456          */
/*   [4] = Scharf, Tausk, Nucl. Phys. B 412 (1994) 423           */
/*   [5] = Broadhurst, Z. Phys. C 47 (1990) 115                  */
/* ------------------------------------------------------------- */

#include "T345.h"
#include "li.h"
#include "qagutil.h"
#include "B0.h"
#include "tinline.h"

#define Errstream cerr

/******************************************************************/
/* Throughout this file the numbers are defined as Double or Cplx */
/* to enable a generalization to arbitrary precision              */
/******************************************************************/

/******************************************************************/
/* convention: vars is frequently used for (var)^2                */
/* for example: m1s denotes m_1^2.                                */
/******************************************************************/

/******************************************************************/
/* some constants which define from which value on the asymptotic */
/* limit of some functions will be taken:                         */
/******************************************************************/

#ifdef QUADPRECISION
static doubledouble ASYMP_LIMIT_DC01=SDouble("1e13");
static doubledouble ASYMP_LIMIT_T3K3_1=SDouble("1e-17");
#else
#define ASYMP_LIMIT_DC01 1e3
#define ASYMP_LIMIT_T3K3_1 1e-7
#endif

#define MAX_BERNOULLI 30

// Bernoulli numbers:
static Double bernoulli[MAX_BERNOULLI+1] =
{ SDouble("1."),
  SDouble("-0.5"),SDouble("0.1666666666666666666666666666667"),SDouble("0"),
  SDouble("-0.0333333333333333333333333333333"),SDouble("0"),
  SDouble("0.02380952380952380952380952380952"),SDouble("0"),
  SDouble("-0.0333333333333333333333333333333"),SDouble("0"),
  SDouble("0.0757575757575757575757575757576"),SDouble("0"),
  SDouble("-0.2531135531135531135531135531136"),SDouble("0"),
  SDouble("1.166666666666666666666666666667"),SDouble("0"),
  SDouble("-7.09215686274509803921568627451"),SDouble("0"),
  SDouble("54.9711779448621553884711779449"),SDouble("0"),
  SDouble("-529.124242424242424242424242424"),SDouble("0"),
  SDouble("6192.12318840579710144927536232"),SDouble("0"),
  SDouble("-86580.2531135531135531135531136"),SDouble("0"),
  SDouble("1.425517166666666666666666666667e6"),SDouble("0"),
  SDouble("-2.729823106781609195402298850575e7"),SDouble("0"),
  SDouble("6.01580873900642368384303868175e8") };

// zeta(n):
#define MAX_ZETA 5
static Double zeta[MAX_ZETA+1] =
{ SDouble("0"),SDouble(" 0"),
  SDouble(" 1.64493406684822643647241516665"),
  SDouble("1.202056903159594285399738161511"),
  SDouble(" 1.08232323371113819151600369654"),
  SDouble("1.036927755143369926331365486457") };

#define Zeta zeta

// faculty n!
#define MAX_FAC 30
static Double fac[MAX_FAC+1] = 
{ SDouble("1."),
  SDouble("1."),
  SDouble("2."),
  SDouble("6."),
  SDouble("24."),
  SDouble("120."),
  SDouble("720."),
  SDouble("5040."),
  SDouble("40320."),
  SDouble("362880."),
  SDouble("3.6288e6"),
  SDouble("3.99168e7"),
  SDouble("4.790016e8"),
  SDouble("6.2270208e9"),
  SDouble("8.71782912e10"),
  SDouble("1.307674368e12"),
  SDouble("2.0922789888e13"),
  SDouble("3.55687428096e14"),
  SDouble("6.402373705728e15"),
  SDouble("1.21645100408832e17"),
  SDouble("2.43290200817664e18"),
  SDouble("5.109094217170944e19"),
  SDouble("1.12400072777760768e21"),
  SDouble("2.585201673888497664e22"),
  SDouble("6.2044840173323943936e23"),
  SDouble("1.5511210043330985984e25"),
  SDouble("4.03291461126605635584e26"),
  SDouble("1.0888869450418352160768e28"),
  SDouble("3.04888344611713860501504e29"),
  SDouble("8.841761993739701954543616e30"),
  SDouble("2.6525285981219105863630848e32")};

inline int even(int p) {if ((p%2)==0) return(TRUE); else return(FALSE); }

static Double ps,m1s,m2s,m3s,m4s,m5s;
static Double prec,abserr;
static Double threshold,tthr1,tthr2;
static Cplx tthr1a,tthr1b;
static Double Rsla235;
static Cplx Csla235;
static Cplx B0fix;
static Cplx T2a(void);
static Cplx T2b(void);
static Cplx T3a(void);
static Cplx T3b(void);
static int T3a_status(void);
static void increase_error(Double err);
static Cplx T3_residuum();
static Cplx T3_residuum_m3zero();
static Cplx T12345integration(Double pse, Double m1se, Double m2se,
                              Double m3se, Double m4se, Double m5se,
                              Double prece, Double *abserre);
static Cplx T2ak1(Double t);  /* integration kernel for T2a  */
static Cplx T2ak2(Double t);  /* integration kernel for T2a  */
static Cplx T2ak3(Double s);  /* integration kernel for T2a  */
static Cplx T2ak4(Double s);  /* integration kernel for T2a  */
static Cplx T2bk1(Double t);  /* integration kernel for T2b  */
static Cplx T2bk2(Double t);  /* integration kernel for T2b  */
static Cplx T2bk3(Double t);  /* integration kernel for T2b  */
static Cplx T2bk4(Double t);  /* integration kernel for T2b  */
static Cplx T3k1(Double t);  /* integration kernel for T3a  */
static Cplx T3k2(Double t);  /* integration kernel for T3a  */
static Cplx T3k3(Double t);  /* integration kernel for T3a  */
static Cplx T3k4(Double t);  /* integration kernel for T3a  */
static Cplx T3k11(Double t);  /* integration kernel for T3a  */
static Cplx T3k12(Double t);  /* integration kernel for T3a  */
static Cplx T3k21(Double t);  /* integration kernel for T3a  */
static Cplx T3k22(Double t);  /* integration kernel for T3a  */
static Cplx T3k31(Double t);  /* integration kernel for T3a  */
static Cplx T3k41(Double t);  /* integration kernel for T3a  */
static Cplx T3k51(Double t);  /* integration kernel for T3a  */

// routines about the discontinuity of the one-loop triangle diagram,
// the mathematical
// background of the according algorithms is described in ref. [2]

#define p1s C0_p1s
#define p2s C0_p2s
#define p3s C0_p3s
#define m1s C0_m1s
#define m2s C0_m2s
#define m3s C0_m3s
#define m1 C0_m1
#define m2 C0_m2
#define m3 C0_m3
#define s0 C0_s0
#define a1 C0_a1
#define a2 C0_a2
#define Deltaps C0_Deltaps
#define Sumps C0_Sumps
#define status C0_status
#define s1_is_real C0_s1_is_real
#define Rs1 C0_Rs1
#define Rs2 C0_Rs2
#define Cs1 C0_Cs1

static Double p1s,p2s,p3s,m1s,m2s,m3s;
static Double m1,m2,m3,s0;
static Double a1,a2;  // for the fast calculation of a
static Double Deltaps,Sumps;
static int status;
static int C0_calc_status(void);
static Cplx C0an_intR(Double s);
static Cplx C0an_intC(Cplx s);
static int s1_is_real;
static Double Rs1,Rs2;
static Cplx Cs1;
static void C0dispersion_init(Double p2se,Double p3se,
			      Double m1se,Double m2se,Double m3se);
static int C0an_is_vanishing;
static Cplx C0_integrate(Double p1se, Double precision);
static Cplx C0_discontinuity(Double s);
static Cplx C0_discontinuitythr(void);
static Cplx C0an(Double p1se);
static Cplx C0k1(Double s);
static Cplx C0k2(Double s);


/*********************************************************************/
/* all what is needed to calculate the discontinuity of the one-loop */
/* three-point function C0, according to ref. [2]:                   */
/*********************************************************************/

static void C0dispersion_init(
    Double p2se,Double p3se,Double m1se,Double m2se,Double m3se)
{
  p2s=p2se; p3s=p3se; m1s=m1se; m2s=m2se; m3s=m3se;
  m1=sqrt(m1s); m2=sqrt(m2s); m3=sqrt(m3s);
  s0=sqr(m1+m2);
  a1=-p2s-p3s+m3s+m3s-m1s-m2s;
  a2=(m1s-m2s)*(p3s-p2s);
  Deltaps=sqr(sqrt(p2s)-sqrt(p3s));
  Sumps=sqr(sqrt(p2s)+sqrt(p3s));
  status=C0_calc_status();
  if ((status==2)||(status==3)||(status==6))
    C0an_is_vanishing=FALSE;
  else
    C0an_is_vanishing=TRUE;
}

// C0_calc_status returns the case the mass configuration belongs to    */
// according to [2]                                                  */

static int C0_calc_status(void)
{
  Double m3sthr;  // threshold value for m3s for the occurrence
                // of the anomalous threshold
  if (m1+m2==Double(0)) m3sthr=m3s+1;   // in fact m3sthr -> Infinity
  else m3sthr=(m1*p2s+m2*p3s)/(m1+m2)-m1*m2;
  if (m3s==Double(0)) {
    if (m3sthr<0) return(11);
    else return(12); };
  Double dm13s,sm13s,dm23s,sm23s;
  dm13s=sqr(m1-m3); sm13s=sqr(m1+m3); dm23s=sqr(m2-m3); sm23s=sqr(m2+m3);
  Cplx l3,l4;
  int l3_is_real,l4_is_real;
  if (p2s>sm23s) {l3_is_real=TRUE; l3=sla(p2s,m2s,m3s);}
  else if (p2s>dm23s) {l3_is_real=FALSE; l3=I*snla(p2s,m2s,m3s);}
       else {l3_is_real=TRUE; l3=-sla(p2s,m2s,m3s);};
  if (p3s>sm13s) {l4_is_real=TRUE; l4=sla(p3s,m1s,m3s);}
  else if (p3s>dm13s) {l4_is_real=FALSE; l4=I*snla(p3s,m1s,m3s);}
       else {l4_is_real=TRUE; l4=-sla(p3s,m1s,m3s);};
  if (l3_is_real==l4_is_real)
  {
    s1_is_real=TRUE;
    Rs1=(-m3s+p2s+p3s+m1s+m2s)*.5-((p2s-m2s)*(p3s-m1s)+real(l3*l4))*0.5/m3s;
    Rs2=(-m3s+p2s+p3s+m1s+m2s)*.5-((p2s-m2s)*(p3s-m1s)-real(l3*l4))*0.5/m3s;
  }
  else
  {
    s1_is_real=FALSE;
    Cs1=(-m3s+p2s+p3s+m1s+m2s)*0.5-((p2s-m2s)*(p3s-m1s)+l3*l4)*0.5/m3s;
  };
  if (m3s >= m3sthr) {return(1);}
  if (la(p2s,m2s,m3sthr)<0)
  {
    if (s1_is_real==TRUE)
    {
      if (Rs1<s0) return(2);
      else return(4);
    }
    else
    {
      return(3);
    };
  }
  else
  {
    if (s1_is_real==TRUE) return(5);
    else return(6);
  };
  Errstream << "Unknown status for the triangle discontinuity.\n";
  return(0);
}

// C0dispersion::discontinuity returns after a proper initialization */
// with p2s,p3s,m1s,m2s,m3s the discontinuity of the triangle        */
// diagram. It is assumed that s=p1s>(sqrt(m1s)+sqrt(m2s))^2

// TO IMPROVE:
//   there are many possibilities to improve the numerics in this routine
//   mainly: only one logarithm would be needed for each evaluation
static Cplx C0_discontinuity(Double s)
{
  Double a,b,l1,l2;
  Cplx Cb,Cl1;
  int b_is_complex; // indicates whether b is complex
  a=s*(s+a1)+a2;
  l2=sla(s,m1s,m2s);
  if (s<Deltaps) {b_is_complex=FALSE; l1=sla(s,p2s,p3s); b=l1*l2;}
  else if (s<Sumps) {b_is_complex=TRUE; Cl1=I*snla(s,p2s,p3s); Cb=Cl1*l2;}
  else  {b_is_complex=FALSE; l1=-sla(s,p2s,p3s); b=l1*l2;};
  if (status==1) {
    if (b_is_complex==FALSE)
//      return(-(TwoPiI/l1)*(clog2(a+b)-clog2(a-b)));  
      return(-(TwoPiI/l1)*(clog2(a+b)-clog1(a-b)));
    else
      return(-(TwoPiI/Cl1)*(log(a+Cb)-log(a-Cb))); };
  if (status==11) {
    if (b_is_complex==FALSE)
      if (s > ASYMP_LIMIT_DC01*(p2s+p3s+m1s+m2s)) {
                       // asymptotic formulae for large s:
        if ((p2s==m2s)&&(p3s!=m1s))
          return(-(TwoPiI/l1)*(clog1(2*p2s*(m1s-p3s)/s)-clog2(a-b)));
        if ((p2s!=m2s)&&(p3s==m1s))
          return(-(TwoPiI/l1)*(clog1(2*p3s*(m2s-p2s)/s)-clog2(a-b)));
        return(-(TwoPiI/l1)*(clog2(2.*(m1s-p3s)*(m2s-p2s))-clog2(a-b))); }
      else
        return(-(TwoPiI/l1)*(clog2(a+b)-clog2(a-b)));
    else
      return(-(TwoPiI/Cl1)*(log(a+Cb)-log(a-Cb))); };
  if (status==12) {   // like status==4, but with s1=+infinity
    Cplx z1=-TwoPiI;
    if (b_is_complex==FALSE)
      if (s > ASYMP_LIMIT_DC01*(p2s+p3s+m1s+m2s)) { // asymptotic formulae:
        if ((p2s==m2s)&&(p3s!=m1s))
          return(-(TwoPiI/l1)*(clog1(2*p2s*(m1s-p3s)/s)-clog2(a-b)+z1));
        if ((p2s!=m2s)&&(p3s==m1s))
          return(-(TwoPiI/l1)*(clog1(2*p3s*(m2s-p2s)/s)-clog2(a-b)+z1));
        return(-(TwoPiI/l1)*(clog1(2.*(m1s-p3s)*(m2s-p2s))-clog2(a-b)+z1)); }
      else
        return(-(TwoPiI/l1)*(clog1(a+b)-clog2(a-b)+z1));
    else
      return(-(TwoPiI/Cl1)*(log(a+Cb)-log(a-Cb)+z1)); };
  if ((status==2)||(status==3)||(status==6)) {
    if (b_is_complex==FALSE)
      return(-(TwoPiI/l1)*(clog1(a+b)-clog2(a-b)));
    else
      return(-(TwoPiI/Cl1)*(log(a+Cb)-log(a-Cb))); };
  if (status==5) {
    if (s<Rs1) {
      if (b_is_complex==FALSE) {
        return(-(TwoPiI/l1)*(clog2(a+b)-clog2(a-b))); }
      else
        return(-(TwoPiI/Cl1)*(log(a+Cb)-log(a-Cb)-TwoPiI)); }
    else {
      if (b_is_complex==FALSE)
        return(-(TwoPiI/l1)*(clog1(a+b)-clog2(a-b)));
      else
        return(-(TwoPiI/Cl1)*(log(a+Cb)-log(a-Cb)));}; }
  if (status==4) {
    Cplx z1;
    if (s<Rs1) z1=-TwoPiI; else z1=Cplx(0);
    if (b_is_complex==FALSE)
      return(-(TwoPiI/l1)*(clog1(a+b)-clog2(a-b)+z1));
    else
      return(-(TwoPiI/Cl1)*(log(a+Cb)-log(a-Cb)+z1)); };
  return(Cplx(0));
}

static Cplx C0_discontinuitythr(void)   // at s=threshold
{
  Double a;
  Cplx Cl1,z1,z2;
  int l1_is_complex; // indicates whether l1 is complex
  Double s=sqr(sqrt(m1s)+sqrt(m2s));
  a=s*(s+a1)+a2;
  if (s<Deltaps) {l1_is_complex=FALSE; Cl1=sla(s,p2s,p3s); }
  else if (s<Sumps) {l1_is_complex=TRUE; Cl1=I*snla(s,p2s,p3s);}
  else  {l1_is_complex=FALSE; Cl1=-sla(s,p2s,p3s);};
  if ((status==1)||(status==11)) return(Cplx(0));
  if (status==12) {   // like status==4, but with s1=+infinity
    if (a<0) return(Cplx(0));
    else return(sqr(TwoPiI)/Cl1); };
  if ((status==2)||(status==3)||(status==6))
    if (a<0) return(-sqr(TwoPiI)/Cl1);
    else return(Cplx(0));
  if (status==5) {
    if (a<0) z2=TwoPiI; else z2=Cplx(0);
    if (s<Rs1) {
      if (l1_is_complex==FALSE) return(Cplx(0));
      else return(-TwoPiI*(z2-TwoPiI)/Cl1); }
    else {
      if (a<0) return(-sqr(TwoPiI)/Cl1); else return(Cplx(0)); } };
  if (status==4) {
    if (s<Rs1) z1=-TwoPiI; else z1=Cplx(0);
    if (a<0) z2=TwoPiI; else z2=Cplx(0);
    return(-(TwoPiI/Cl1)*(z2+z1)); };
  return(Cplx(0));
}


// for the calculation of the contribution of the anomalous threshold ([2])
static Cplx C0an_intR(Double s)
{
  Double l6=sla(p1s,p2s,p3s);
  int l5_is_real;
  Double Rl5;
  Cplx Cl5;
  if (s<Deltaps) {l5_is_real=TRUE; Rl5=sla(s,p2s,p3s);}
  else {
    if (s>Sumps) {l5_is_real=TRUE; Rl5=-sla(s,p2s,p3s);}
    else {l5_is_real=FALSE; Cl5=I*snla(s,p2s,p3s);};
  };
  if (l5_is_real==TRUE)
    return((clog1(p1s-s)
            -clog1(-s*(p2s+p3s-p1s)+sqr(p2s-p3s)-p1s*(p2s+p3s)+Rl5*l6))/l6);
  else
    return((clog1(p1s-s)
            -log(-s*(p2s+p3s-p1s)+sqr(p2s-p3s)-p1s*(p2s+p3s)+Cl5*l6))/l6);
}

// the same, but for complex s1:
static Cplx C0an_intC(Cplx s)
{
  Double l6=sla(p1s,p2s,p3s);
  Cplx Cl5=sqrt(Deltaps-s)*sqrt(Sumps-s);
  Cplx r1=-s*(p2s+p3s-p1s)+sqr(p2s-p3s)-p1s*(p2s+p3s)+Cl5*l6;
  if (imag(r1)<0)
    return((log(p1s-s)-log(r1)-TwoPiI)/l6);
  else
    return((log(p1s-s)-log(r1))/l6);
}


// this is now the total contribution
static Cplx C0an(Double p1se)
{
  p1s=p1se;
  if (s1_is_real==TRUE) return(TwoPiI*(C0an_intR(Rs1)-C0an_intR(s0)));
  else return(TwoPiI*(C0an_intC(Cs1)-C0an_intR(s0)));
}


// this routine returns the C0-function, for debugging purposes:
static Cplx C0_integrate(Double p1se, Double prec)
{
  Cplx C0an_result=Cplx(0);
  if (C0an_is_vanishing==FALSE) C0an_result=C0an(p1se);
  p1s=p1se;
  Double dummy;
  return(Cqagpinf(C0_discontinuity,s0,p1s,1,prec,&dummy)/TwoPiI+C0an_result);
}

#undef p1s
#undef p2s
#undef p3s
#undef m1s
#undef m2s
#undef m3s
#undef m1
#undef m2
#undef m3
#undef s0
#undef a1
#undef a2
#undef Deltaps
#undef Sumps
#undef status
#undef s1_is_real
#undef Rs1
#undef Rs2
#undef Cs1


/*********************************************************************/
/* The two-loop master integral,                                     */
/* according to ref. [2] and [5]:                                    */
/*********************************************************************/

inline bool BelowThreshold2(Double ps,Double m1s,Double m2s)
{ return(ps<sqr(sqrt(m1s)+sqrt(m2s))); }

inline bool BelowThreshold3(Double ps,Double m1s,Double m2s,Double m3s)
{ return(ps<sqr(sqrt(m1s)+sqrt(m2s)+sqrt(m3s))); }

static Cplx T12345integration(Double pse,Double m1se,Double m2se,
			      Double m3se,Double m4se,
			      Double m5se,Double prece,Double *abserre)
{
  ps=pse; m1s=m1se; m2s=m2se; m3s=m3se; m4s=m4se; m5s=m5se; prec=prece;
  
  abserr=0;
  Cplx z=T2a()+T2b()+T3a()+T3b();
  *abserre=abserr;
  if (  BelowThreshold2(pse,m1s,m2s)&&BelowThreshold2(pse,m4s,m5s)
      &&BelowThreshold3(pse,m1s,m3s,m5s)&&BelowThreshold3(pse,m2s,m3s,m4s))
    z=real(z);
  return(z);
}


static Cplx T2ak1(Double t)
{
  return(conj(C0_discontinuity(t))*(B0A(t,m1s,m2s)-B0fix)/(t-ps));
}

static Cplx T2ak2(Double t)
{
  return(conj(C0_discontinuity(t))*(B0A(t,m1s,m2s)-B0fix));
}


static Cplx T2ak3(Double s)
{
  return(TwoPiI*sla(s,m1s,m2s)*conj(C0an(s))/(s*(s-ps)));
}

static Cplx T2ak4(Double s)
{
  return(TwoPiI*sla(s,m1s,m2s)*conj(C0an(s))/s);
}


static Cplx T2bk1(Double t)
{
  return(C0_discontinuity(t)*(B0fix-conj(B0A(t,m4s,m5s)))/(t-ps));
}

static Cplx T2bk2(Double t)
{
  return(C0_discontinuity(t)*(B0fix-conj(B0A(t,m4s,m5s))));
}

static Cplx T2bk3(Double s)
{
  return(TwoPiI*sla(s,m4s,m5s)*C0an(s)/(s*(s-ps)));
}

static Cplx T2bk4(Double s)
{
  return(TwoPiI*sla(s,m4s,m5s)*C0an(s)/s);
}


// the two-particle contributions T2a and T2b (see [2]) are very
// similar. It simplifies therefore the code developpment to define the masses
// with aliases:

#define M1S m1s
#define M2S m2s
#define M4S m4s
#define M5S m5s
#define TM21 T2ak1
#define TM22 T2ak2
#define TM23 T2ak3
#define TM24 T2ak4

static Cplx T2a(void)
{
  C0dispersion_init(M1S,M2S,M5S,M4S,m3s);
  B0fix=B0A(ps,M1S,M2S);
  Cplx z1,z2;
  Double threshold1=sqr(sqrt(M4S)+sqrt(M5S));
  Double err1=0,err2=0;
  Double M1=sqrt(M1S),M2=sqrt(M1S);
  Double lim1=sqr(M1-M2),lim2=sqr(M1+M2);
  z1=Cqagpinf(TM22,threshold1,ps,1,prec,&err1,lim1,lim2);
  increase_error(abs(err1/TwoPiI));
  if (C0an_is_vanishing==TRUE) z2=Cplx(0);
  else {
    err1=0; err2=0;
    Double threshold2=sqr(sqrt(M1S)+sqrt(M2S));
    z2=Cqagpinf(TM24,threshold2,ps,1,prec,&err1);
    increase_error(abs(err2/TwoPiI));
  };
  return((z1+z2)/TwoPiI);
}


#undef M1S
#undef M2S
#undef M4S
#undef M5S
#undef TM21
#undef TM22
#undef TM23
#undef TM24


#define M1S m4s
#define M2S m5s
#define M4S m1s
#define M5S m2s
#define TM21 T2bk1
#define TM22 T2bk2
#define TM23 T2bk3
#define TM24 T2bk4

static Cplx T2b(void)
{
  C0dispersion_init(M1S,M2S,M5S,M4S,m3s);
  B0fix=B0A(ps,M1S,M2S);
  Cplx z1,z2;
  Double threshold1=sqr(sqrt(M4S)+sqrt(M5S));
  Double err1=0,err2=0;
  Double M1=sqrt(M1S),M2=sqrt(M1S);
  Double lim1=sqr(M1-M2),lim2=sqr(M1+M2);
  z1=Cqagpinf(TM22,threshold1,ps,1,prec,&err1,lim1,lim2);
  increase_error(abs(err1/TwoPiI));
  if (C0an_is_vanishing==TRUE) z2=Cplx(0);
  else {
    err1=0; err2=0;
    Double threshold2=sqr(sqrt(M1S)+sqrt(M2S));
    z2=Cqagpinf(TM24,threshold2,ps,1,prec,&err1);
    increase_error(abs(err2/TwoPiI));
  };
  return((z1+z2)/TwoPiI);
}

#undef M1S
#undef M2S
#undef M4S
#undef M5S
#undef TM21
#undef TM22
#undef TM23
#undef TM24


// the difference of two B0-functions, divided by the difference of
// the arguments:
// TO IMPROVE:  asymptotic formula for pse near to si
static Cplx DiffB0R(Double pse,Double si,Double te,Double m2se)
{
  return((B0A(pse,te,m2se)-B0A(si,te,m2se))/(si-pse));
}

inline bool approximately_equal(Double x1, Double x2, Double scale)
{ return(abs((x1-x2)/(abs(x1)+abs(x2))) < scale); }

static Cplx DiffB0Rs12(Double pse,Double s12,Double te,Double m2se)
// approximation of ((DiffB0R(ps,s1)-DiffB0R(ps,s2))/(s1-s2)) for s1 ~ s2
{
  return((B0A(s12,te,m2se)-B0A(pse,te,m2se))/sqr(s12-pse)
         -DB0(s12,te,m2se)/(s12-pse));
}

/*
#define DiffB0ASYMP_LIMIT .00001

static Cplx DiffB0R(Double pse,Double si,Double te,Double m2se)
{
  if (pse !=Double(0))
    if (abs((pse-si)/pse)<DiffB0ASYMP_LIMIT) 
      return(-DB0((pse+si)/2,te,m2se));
  return((B0A(pse,te,m2se)-B0A(si,te,m2se))/(si-pse));
}
*/

// for B0(si-i*eps):
Cplx DiffB0R_cc(Double pse,Double si,Double te,Double m2se)
{
  return((B0A(pse,te,m2se)-conj(B0A(si,te,m2se)))/(si-pse));
}


// for complex si:
Cplx DiffB0C(Double pse,Cplx si,Double te,Double m2se)
{
  return((B0A(pse,te,m2se)-B0A(si,te,m2se))/(si-pse));
}


static Cplx T3k1(Double t)  // main case
{
  Double m2=sqrt(m2s);
  Double st=sqrt(t);
  Double la1=la(t,m3s,m4s);
  Double sla1=sqrt(la1);
  Double r1=t*(2*m2s-m5s+m3s)-m2s*(m4s-m3s);
  Double r2=(t-m2s)*(t*(m5s-m3s)-m2s*(m4s-m3s));
  Double r3=m3s*(m3s-t-m4s)/la1;
  Double r4=m3s
           +( t*(2*m4s*m3s-m4s*m2s+m5s*m4s+m5s*m3s)
             +m4s*(m2s-m5s)*(m4s-m3s)
             +m5s*m3s*(m4s-m3s))
            /la1;
  Double x1=0.5*(t+m2s+m4s+m5s-m3s+(t-m4s)*(m5s-m2s)/m3s);
  Double x2=0.5*sla1*Rsla235/m3s;
  Double s1=x1+x2;
  Double s2=x1-x2;
  Double s3=sqr(st+m2);
  Double s4=sqr(st-m2);
  Double f1=(s1*r1+r2)/((s1-s2)*(s1-s3)*(s1-s4))+(s1*r3+r4)/(s1-s2);
  Double f2=(s2*r1+r2)/((s2-s1)*(s2-s3)*(s2-s4))+(s2*r3+r4)/(s2-s1);
  Double f3=(s3*r1+r2)/((s3-s1)*(s3-s2)*(s3-s4));
  Double f4=(s4*r1+r2)/((s4-s1)*(s4-s2)*(s4-s3));
  Cplx z1;
  if (m1s==Double(0)) z1=log(t)*sla1/(t*m3s);
  else z1=clog1(t/m1s-1.)*sla1/(t*m3s);
  return(z1*(f1*DiffB0R(ps,s1,t,m2s)+f2*DiffB0R(ps,s2,t,m2s)+
             f3*DiffB0R(ps,s3,t,m2s)+f4*DiffB0R(ps,s4,t,m2s)));
}


static Cplx T3k61(Double t)  // sla235=0
{
  Double m2=sqrt(m2s);
  Double st=sqrt(t);
  Double la1=la(t,m3s,m4s);
  Double sla1=sqrt(la1);
  Double r1=t*(2*m2s-m5s+m3s)-m2s*(m4s-m3s);
  Double r2=(t-m2s)*(t*(m5s-m3s)-m2s*(m4s-m3s));
  Double r3=m3s*(m3s-t-m4s)/la1;
  Double r4=m3s
           +( t*(2*m4s*m3s-m4s*m2s+m5s*m4s+m5s*m3s)
             +m4s*(m2s-m5s)*(m4s-m3s)
             +m5s*m3s*(m4s-m3s))
            /la1;
  Double s1=0.5*(t+m2s+m4s+m5s-m3s+(t-m4s)*(m5s-m2s)/m3s);
  Double s3=sqr(st+m2);
  Double s4=sqr(st-m2);
  Cplx z1;
  if (m1s==Double(0)) z1=log(t)*sla1/(t*m3s);
  else z1=clog1(t/m1s-1.)*sla1/(t*m3s);
  if (approximately_equal(s1,ps,ASYMP_LIMIT_T3K3_1)) {
    // on-shell diagram with ps=m4s, m5s=0, m2s=m3s
    Double f11=-r3+(r2*(2.*s1-s3-s4)+r1*(s1*s1-s3*s4))/sqr(s1-s3)/sqr(s1-s4);
    Double f12=-r3*s1-r4+(r1*s1+r2)/(s1-s3)/(s4-s1);
    Double f3=(r1*s3+r2)/powint(s3-s1,3)/(s4-s3);
    Double f4=(r1*s4+r2)/powint(s4-s1,3)/(s3-s4);
    Double f1=-f3-f4;
    // taking only the real part is crude and requires further studies: 
    return(z1*(f12*DDB0(s1,t,m2s)+f11*DB0(s1,t,m2s)+f1*B0A(s1,t,m2s)
	            +f3*B0A(s3,t,m2s)+f4*B0A(s4,t,m2s))); }
  else {
  Double f11=(r3*s1+r4+(r1*s1+r2)/(s1-s3)/(s1-s4))/(ps-s1);
  Double f1=-(r3*s1+r4+(r1*s1+r2)/(s1-s3)/(s1-s4))/(ps-s1)
       +r3-(r2*(2.*s1-s3-s4)+r1*(s1*s1-s3*s4))/sqr(s1-s3)/sqr(s1-s4);
  Double f3=(s3*r1+r2)/(sqr(s3-s1)*(s3-s4));
  Double f4=(s4*r1+r2)/(sqr(s4-s1)*(s4-s3));
  return(z1*(f1*DiffB0R(ps,s1,t,m2s)+f11*DB0(s1,t,m2s)+
             f3*DiffB0R(ps,s3,t,m2s)+f4*DiffB0R(ps,s4,t,m2s))); };
}

static Cplx T3k2(Double t)  // for imaginary sla(m2s,m3s,m5s)
{
  Double m2=sqrt(m2s);
  Double st=sqrt(t);
  Double la1=la(t,m3s,m4s);
  Double sla1=sqrt(la1);
  Double r1=t*(2*m2s-m5s+m3s)-m2s*(m4s-m3s);
  Double r2=(t-m2s)*(t*(m5s-m3s)-m2s*(m4s-m3s));
  Double r3=m3s*(m3s-t-m4s)/la1;
  Double r4=m3s
           +( t*(2*m4s*m3s-m4s*m2s+m5s*m4s+m5s*m3s)
             +m4s*(m2s-m5s)*(m4s-m3s)
             +m5s*m3s*(m4s-m3s))
            /la1;
  Double x1=0.5*(t+m2s+m4s+m5s-m3s+(t-m4s)*(m5s-m2s)/m3s);
  Cplx x2=0.5*sla1*Csla235/m3s;
  Cplx s1=x1+x2;
  Cplx s2=conj(s1);
  Double s3=sqr(st+m2);
  Double s4=sqr(st-m2);
  Cplx f1=(s1*r1+r2)/(2.*I*imag(s1)*(s1-s3)*(s1-s4))
           +(s1*r3+r4)/(2.*I*imag(s1));
  Cplx f2=conj(f1);
  Cplx f3=(s3*r1+r2)/((s3-s1)*(s3-conj(s1))*(s3-s4));
  Cplx f4=(s4*r1+r2)/((s4-s1)*(s4-conj(s1))*(s4-s3));
  Cplx z1;
  if (m1s==Double(0)) z1=log(t)*sla1/(t*m3s);
  else z1=clog1(t/m1s-1.)*sla1/(t*m3s);
  return(z1*(f1*DiffB0C(ps,s1,t,m2s)+f2*DiffB0C(ps,s2,t,m2s)+
             f3*DiffB0R(ps,s3,t,m2s)+f4*DiffB0R(ps,s4,t,m2s)));
}

static Cplx T3k3(Double t)  // with threshold t_thr1 ([2])
{
  Double m2=sqrt(m2s);
  Double st=sqrt(t);
  Double la1=la(t,m3s,m4s);
  Double sla1=sqrt(la1);
  Double r1=t*(2*m2s-m5s+m3s)-m2s*(m4s-m3s);
  Double r2=(t-m2s)*(t*(m5s-m3s)-m2s*(m4s-m3s));
  Double r3=m3s*(m3s-t-m4s)/la1;
  Double r4=m3s
           +( t*(2*m4s*m3s-m4s*m2s+m5s*m4s+m5s*m3s)
             +m4s*(m2s-m5s)*(m4s-m3s)
             +m5s*m3s*(m4s-m3s))
            /la1;
  Double x1=0.5*(t+m2s+m4s+m5s-m3s+(t-m4s)*(m5s-m2s)/m3s);
  Double x2=0.5*sla1*Rsla235/m3s;
  Double s1=x1+x2;
  Double s2=x1-x2;
  Double s3=sqr(st+m2);
  Double s4=sqr(st-m2);
  Double f1=(s1*r1+r2)/((s1-s2)*(s1-s3)*(s1-s4))+(s1*r3+r4)/(s1-s2);
  Double f2=(s2*r1+r2)/((s2-s1)*(s2-s3)*(s2-s4))+(s2*r3+r4)/(s2-s1);
  Double f3=(s3*r1+r2)/((s3-s1)*(s3-s2)*(s3-s4));
  Double f4=(s4*r1+r2)/((s4-s1)*(s4-s2)*(s4-s3));
  Cplx z1;
  if (m1s==Double(0)) z1=log(t)*sla1/(t*m3s);
  else z1=clog1(t/m1s-1.)*sla1/(t*m3s);
  if (t>tthr1)
    return(z1*(f1*DiffB0R(ps,s1,t,m2s)+f2*DiffB0R_cc(ps,s2,t,m2s)+
               f3*DiffB0R(ps,s3,t,m2s)+f4*DiffB0R(ps,s4,t,m2s)));
  else {
    if (approximately_equal(s1,s2,ASYMP_LIMIT_T3K3_1)==TRUE) {
       Double s12=(s1+s2)/2;
       Double f12=(s12*r1+r2)/((s12-s3)*(s12-s4))+(s12*r3+r4);
       return(z1*(f12*DiffB0Rs12(ps,s12,t,m2s)
                  +(  r3+r1/((s12 - s3)*(s12 - s4))
                      -(r2 + r1*s12)/((s12 - s4)*sqr(s12 - s3))
                      -(r2 + r1*s12)/((s12 - s3)*sqr(s12 - s4)))
                    *DiffB0R(ps,s12,t,m2s)
                  +f3*DiffB0R(ps,s3,t,m2s)+f4*DiffB0R(ps,s4,t,m2s))); }
    else
      return(z1*(f1*DiffB0R(ps,s1,t,m2s)+f2*DiffB0R(ps,s2,t,m2s)+
                 f3*DiffB0R(ps,s3,t,m2s)+f4*DiffB0R(ps,s4,t,m2s))); };
}

static Cplx T3k4(Double t)  // principal value with thresholds t_thr1,t_thr2
{
  return(T3k3(t)+T3k3(2.*tthr2-t));
}

static Cplx T3k11(Double t)  // for m3=0
{
  Double m2=sqrt(m2s);
  Double st=sqrt(t);
  Double s1=sqr(st+m2);
  Double s2=sqr(st-m2);
  Double c1=(m2s-m5s)*(t-m4s);
  Double c2=-t*(m5s*t+sqr(m5s)-m4s*m5s-m2s*m5s-m2s*m4s)-m2s*m4s*(m2s+m4s-m5s);
  Double s3=c2/c1;
  Double r0=2*(-m2s + t)*(sqr(m2s)*m4s + m2s*sqr(m4s) - m2s*m4s*m5s - 
          2*m2s*m4s*t + m5s*sqr(t));
  Double r1=2*(2*sqr(m2s)*m4s + m2s*sqr(m4s) - 2*m2s*m4s*m5s - m2s*m4s*t - 
          m4s*m5s*t + 2*m2s*sqr(t) - m5s*sqr(t));
  Double r2=2*m4s*(-m2s + m5s);
  Double f1=(r0+r1*s1+r2*sqr(s1))/((s1-s2)*(s1-s3));
  Double f2=(r0+r1*s2+r2*sqr(s2))/((s2-s1)*(s2-s3));
  Double f3=(r0+r1*s3+r2*sqr(s3))/((s3-s1)*(s3-s2));
  Cplx z1=clog1(t/m1s-1.)/(2.*t*c1);
  if (t>tthr1)
    return(z1*(f1*DiffB0R(ps,s1,t,m2s)+f2*DiffB0R(ps,s2,t,m2s)+
               f3*DiffB0R_cc(ps,s3,t,m2s)));
  else
    return(z1*(f1*DiffB0R(ps,s1,t,m2s)+f2*DiffB0R(ps,s2,t,m2s)+
               f3*DiffB0R(ps,s3,t,m2s)));
}


static Cplx T3k21(Double t)  // special case with m1=m3=m5, m2=m4=0
// The formula for this case cannot be derived in the same way as
// in the other cases, since the border term in the partial integration
// does not vanish! But no partial integration is neccessary to obtain
// a integral over logarithms only.
{
  return((clog1(t/(t-ps))/ps)*log(m1s/t)/(t-m1s));
}

static Cplx T3k22(Double t)  // special case with m2=m3=m4, m1=m5=0
{
  Double m=sqrt(m2s);
  Double st=sqrt(t);
  Double s1=sqr(st+m);
  Double s2=sqr(st-m);
  Double z1=0.5*log(t)/(m*t*sqrt(t-4.*m2s));
  return(z1*((t-2.*m2s-st*m)*DiffB0R(ps,s1,t,m2s)
             -(t-2.*m2s+st*m)*DiffB0R(ps,s2,t,m2s)));
}


static Cplx T3k31(Double t)  // special case with m2=m3, m4=m5=0
{
  Double m=sqrt(m2s);
  Double st=sqrt(t);
  Double s1=sqr(st+m);
  Double s2=sqr(st-m);
  Double z1=0.5*log(t/m1s-1)/(m*t*st);
  return(z1*((t-m*st)*DiffB0R(ps,s1,t,m2s)
             -(t+st*m)*DiffB0R(ps,s2,t,m2s)));
}



static Cplx T3k41(Double t)  // special case, m2 != m3 != m1, m4=m5=0
{
  Double m2=sqrt(m2s);
  Double st=sqrt(t);
  Double r0=-((2.*t-m3s)*t-m2s*m3s)*(m3s-m2s);
  Double r1=-m3s*(-3.*t+m3s-2.*m2s);
  Double r2=-m3s;
  Double s1=sqr(st+m2);
  Double s2=sqr(st-m2);
  Double s3=(1.-m2s/m3s)*t+m2s-m3s;
  Double f1=(s1*s1*r2+r1*s1+r0)/((s1-s2)*(s1-s3));
  Double f2=(s2*s2*r2+r1*s2+r0)/((s2-s1)*(s2-s3));
  Double f3=(s3*s3*r2+r1*s3+r0)/((s3-s1)*(s3-s2));
  Cplx z1;
  if (m1s==Double(0)) z1=log(t)/(t*m3s);
  else z1=clog1(t/m1s-1.)/(t*m3s);
  return(z1*(f1*DiffB0R(ps,s1,t,m2s)+f2*DiffB0R(ps,s2,t,m2s)+
             f3*DiffB0R(ps,s3,t,m2s)));
}


static Cplx T3k51(Double t)  // special case, m2=0
{
  Double slatm3m4=sla(t,m3s,m4s);
  Double r0=t*t*(m5s*(m3s+m4s-t)-2.*m3s*m4s);
  Double r1=2.*m3s*t*t+(m3s*(m4s+m5s-3.*m3s)+m4s*m5s)*t+(m3s-m5s)*sqr(m4s-m3s);
  Double r2=m3s*(m3s-m4s-t);
  Double s12=t*(m3s+m5s)+m3s*(m4s+m5s-m3s)-m4s*m5s;
  Double s1=.5*(s12+(m3s-m5s)*slatm3m4)/m3s;
  Double s2=.5*(s12-(m3s-m5s)*slatm3m4)/m3s;
  Double s3=t;
  Double f1=(s1*s1*r2+r1*s1+r0)/((s1-s2)*(s1-s3));
  Double f2=(s2*s2*r2+r1*s2+r0)/((s2-s1)*(s2-s3));
  Double f3=(s3*s3*r2+r1*s3+r0)/((s3-s1)*(s3-s2));
  Cplx z1;
  if (m1s==Double(0)) z1=log(t)/(t*m3s*slatm3m4);
  else z1=clog1(t/m1s-1.)/(t*m3s*slatm3m4);
  return(z1*(f1*DiffB0R(ps,s1,t,m2s)+f2*DiffB0R(ps,s2,t,m2s)+
             f3*DiffB0R(ps,s3,t,m2s)));
}


static Cplx T3k12(Double t)  // principal value with thresholds t_thr1,t_thr2
{
  return(T3k11(t)+T3k11(2.*tthr2-t));
}

#define T3_NORMAL 1           // main case
#define T3_CMPLX 2            // complex sla(m2s,m3s,m5s)
#define T3_THR1 3             // threshold t_thr1
#define T3_THR2 4             // thresholds t_thr1, t_thr2
#define T3_M3ZERO 11          // m3=0
#define T3_M3ZERO_THR2 12     // m3=0, threshold t_thr2
#define T3_M1M3M5 21          // m1=m3=m5, m2=m4=0
#define T3_M2M3M4 22          // m2=m3=m4, m1=m5=0
#define T3_M2M3 31            // m2=m3, m4=5=0
#define T3_M2NEQM3 41         // m2 != m3, m4=m5=0
#define T3_M2ZERO 51          // m2=0
#define T3_LA235Z 61          // la235=0, for example m5=0, m2=m3

static int T3a_status(void)
{
  if ((m5s==0)&&(m2s==m3s)) {
    Rsla235=0;
    return(T3_LA235Z); };
  Double la235=la(m2s,m3s,m5s);
  if (la235==0) {
    Rsla235=0;
    return(T3_LA235Z); };

// for debugging:   
  tthr1a=(la235+2.*m2s*(m3s+m4s)+(m5s-m2s-m3s)*csqrt1(la235+4.*m2s*m4s))
           /(2.*m2s);
  tthr1b=(la235+2.*m2s*(m3s+m4s)-(m5s-m2s-m3s)*csqrt1(la235+4.*m2s*m4s))
           /(2.*m2s);

  Double m2=sqrt(m2s);
  Double m5=sqrt(m5s);
  Double la45p;
  if ((m1s==m3s)&&(m3s==m5s)&&(m2s==Double(0))&&(m4s==Double(0)))
    return(T3_M1M3M5);
  if ((m2s==m3s)&&(m3s==m4s)&&(m1s==Double(0))&&(m5s==Double(0)))
    return(T3_M2M3M4);
  if ((m2s==m3s)&&(m4s==Double(0))&&(m5s==Double(0))) return(T3_M2M3);
  if ((m1s!=m3s)&&(m4s==Double(0))&&(m5s==Double(0))) return(T3_M2NEQM3);
  if (m3s==Double(0)) {
    tthr1=
      (la235+2.*m2s*m4s+(m5s-m2s)*sqrt(la235+4.*m2s*m4s))/(2.*m2s);
//   ??  TO IMPROVE:
//   the following conditions for thr2 seem to be not yet ok in all cases
    la45p=la(ps,m4s,m5s);
    if (la45p<0) return(T3_M3ZERO);
    Rsla235=sqrt(la235);
    tthr2=(m5s*(ps+m2s+m4s-m5s)+m2s*(m4s-ps)+Rsla235*sqrt(la45p))
            /(2*m5s);
    if ((tthr2>threshold)&&(tthr2>tthr1)&&(m5s>m2s))
      return(T3_M3ZERO_THR2);
    else
      return(T3_M3ZERO); };
  if (m2s==Double(0)) {
    Rsla235=m3s-m5s;
    return(T3_M2ZERO); };
  if (m3s>sqr(m2+m5)) {
    Rsla235=sqrt(la235);
    return(T3_NORMAL); };
  if (m3s>sqr(m2-m5)) {
    Csla235=I*sqrt(-la235);
    return(T3_CMPLX); };
  Rsla235=sqrt(la235);
  tthr1=(la235+2.*m2s*(m3s+m4s)+(m5s-m2s-m3s)*sqrt(la235+4.*m2s*m4s))/(2.*m2s);
  la45p=la(ps,m4s,m5s);
  if (la45p<0) return(T3_THR1);
  tthr2=(m5s*(ps+m2s+m3s+m4s-m5s)+(m2s-m3s)*(m4s-ps)+Rsla235*sqrt(la45p))
        /(2*m5s);
  if ((tthr2>threshold)&&(tthr2>tthr1)&&(m3s+m5s-m2s>Rsla235))
    return(T3_THR2);
  else
    return(T3_THR1);
}

// the residuum contribution in case of t_thr2:
static Cplx T3_residuum()
{
  Double m2=sqrt(m2s);
  Double t=tthr2;
  Double st=sqrt(t);
  Double la1=la(t,m3s,m4s);
  Double sla1=sqrt(la1);
  Double r1=t*(2*m2s-m5s+m3s)-m2s*(m4s-m3s);
  Double r2=(t-m2s)*(t*(m5s-m3s)-m2s*(m4s-m3s));
  Double r3=m3s*(m3s-t-m4s)/la1;
  Double r4=m3s
           +( t*(2*m4s*m3s-m4s*m2s+m5s*m4s+m5s*m3s)
             +m4s*(m2s-m5s)*(m4s-m3s)
             +m5s*m3s*(m4s-m3s))
            /la1;
  Double x1=0.5*(t+m2s+m4s+m5s-m3s+(t-m4s)*(m5s-m2s)/m3s);
  Double x2=0.5*sla1*Rsla235/m3s;

  Double s1=x1+x2;
  Double s2=x1-x2;
  Double s3=sqr(st+m2);
  Double s4=sqr(st-m2);
  Double f2=(s2*r1+r2)/((s2-s1)*(s2-s3)*(s2-s4))+(s2*r3+r4)/(s2-s1);
  
  Cplx z1=clog1(t/m1s-1.)*sla1/(t*m3s);
  Double dsdt=(m3s-m2s+m5s+(m3s+m4s-tthr2)*Rsla235/sla(tthr2,m3s,m4s))/(2.*m3s);
  return(PiI*z1*f2*(TwoPiI*sla(ps,tthr2,m2s)/ps)/dsdt);
}

// the same for m3=0:
static Cplx T3_residuum_m3zero()
{
  Double m2=sqrt(m2s);
  Double t=tthr2;
  Double st=sqrt(t);
  Double s1=sqr(st+m2);
  Double s2=sqr(st-m2);
  Double c1=(m2s-m5s)*(t-m4s);
  Double c2=-t*(m5s*t+sqr(m5s)-m4s*m5s-m2s*m5s-m2s*m4s)-m2s*m4s*(m2s+m4s-m5s);
  Double s3=c2/c1;
  Double r0=2*(-m2s + t)*(sqr(m2s)*m4s + m2s*sqr(m4s) - m2s*m4s*m5s - 
          2*m2s*m4s*t + m5s*sqr(t));
  Double r1=2*(2*sqr(m2s)*m4s + m2s*sqr(m4s) - 2*m2s*m4s*m5s - m2s*m4s*t - 
          m4s*m5s*t + 2*m2s*sqr(t) - m5s*sqr(t));
  Double r2=2*m4s*(-m2s + m5s);
  Double f3=(r0+r1*s3+r2*sqr(s3))/((s3-s1)*(s3-s2));
  Cplx z1=clog1(t/m1s-1.)/(2.*t*c1);
  Double dsdt=(-(sqr(m2s)*m4s) + 2*m2s*m4s*m5s + sqr(m4s)*m5s - m4s*sqr(m5s) - 
            2*m4s*m5s*t + m5s*sqr(t))/((-m2s + m5s)*sqr(-m4s + t));
  return(PiI*z1*f3*(TwoPiI*sla(ps,tthr2,m2s)/ps)/dsdt);
}

static Cplx T3a(void)
{
  threshold=sqr(sqrt(m3s)+sqrt(m4s));
  Double err1;
  Double err2=0;
  Cplx z;
  int i=T3a_status();
  switch(i) {
    case T3_NORMAL:  z=Cqaginfasqrtsing(T3k1,threshold,prec,&err1); break;
    case T3_CMPLX :  z=Cqaginfasqrtsing(T3k2,threshold,prec,&err1); break;
    case T3_THR1  :  z=Cqaginfasqrtsing(T3k3,threshold,prec,&err1); break;
    case T3_THR2  :  z=Cqag(T3k4,threshold,tthr2,prec,&err1)
                       +Cqaginf(T3k3,2.*tthr2-threshold,prec,&err2)
                       +T3_residuum(); break;
    case T3_M3ZERO:  z=Cqaginfasqrtsing(T3k11,threshold,prec,&err1); break;
    case T3_M3ZERO_THR2:
                     z=Cqag(T3k12,threshold,tthr2,prec,&err1)
                       +Cqaginf(T3k11,2*tthr2-threshold,prec,&err2)
                       +T3_residuum_m3zero(); break;
    case T3_M1M3M5:  z=Cqaginf(T3k21,threshold,prec,&err1); break;
    case T3_M2M3M4:  z=Cqaginfasqrtsing(T3k22,threshold,prec,&err1); break;
    case T3_M2M3:    z=Cqaginfasqrtsing(T3k31,threshold,prec,&err1); break;
    case T3_M2NEQM3: z=Cqaginfasqrtsing(T3k41,threshold,prec,&err1); break;
    case T3_M2ZERO:  z=Cqaginfasqrtsing(T3k51,threshold,prec,&err1); break;
    case T3_LA235Z:  z=Cqaginfasqrtsing(T3k61,threshold,prec,&err1); break;
//    case T3_LA235Z:  z=Cqaginf(T3k61,threshold,prec,&err1); break;
  }
  increase_error(err1);
  increase_error(err2);
  return(z);
}

static Cplx T3b(void)
{
  interchange(&m1s,&m2s);
  interchange(&m4s,&m5s);
  Cplx z=T3a();
  interchange(&m1s,&m2s);
  interchange(&m4s,&m5s);
  return(z);
}

static void increase_error(Double err)
{
  abserr=sqrt(sqr(err)+sqr(abserr));
}

/******************************************************************/
/* for some cases the formulas of [5] are required. Some helpful  */
/* functions:                                                     */
/* fnx are the functions defined by Broadhurst ([5])              */
/* n stands for the index, x is "p" for F+ and "n" for F-         */
/* for example f3p = F3+                                          */
/******************************************************************/
static Double fb1(int p,int sign)  // for argument 1
{
  if (sign==-1) return((powint(Double(2),1-p)-1)*fb1(p,1));
  if (p<2) {
    if (even(p)==TRUE) return(0);
    else return(p*bernoulli[1-p]); }
  else {
    if (even(p)==TRUE) return(powint(2*Pi,p)*abs(bernoulli[p])/(2*fac[p-2]));
    else return(p*(p-1)*zeta[p]); };
}

static Cplx f3p(Double x)
{
  if (x==1.) return(6*zeta[3]);
  if (x==Double(0)) return(Cplx(0));
  return(-3.*clog2(1-x)*sqr(clog1(x))-6.*clog1(x)*li2(x)-2.*clog1(x)*li2(1-x)
         +6.*li3(x)+PiSonThree*clog1(x));
}


static Cplx f3m(Double x)
{
  if (x==1.) return(-4.5*zeta[3]);
  if (x==Double(0)) return(Cplx(0));
  return(-clog1(1+x)*sqr(clog1(x))-4.*clog1(x)*li2(-x)+6.*li3(-x));
}


static Cplx f3p(Cplx x)
{
  if (x==1.) return(6*zeta[3]);
  if (x==Double(0)) return(Cplx(0));
  return(-3.*log(1-x)*sqr(log(x))-6.*log(x)*li2(x)-2.*log(x)*li2(1-x)
         +6.*li3(x)+PiSonThree*log(x));
}


static Cplx f3m(Cplx x)
{
  if (x==1.) return(-4.5*zeta[3]);
  if (x==Double(0)) return(Cplx(0));
  return(-log(1+x)*sqr(log(x))-4.*log(x)*li2(-x)+6.*li3(-x));
}


// m3s=0, m1s=m2s=m4s=m5s=ms,  ps > 4*ms
static Cplx T12345_special1a(Double ps,Double ms)
{
  Double s=sqrt(1.-4.*ms/ps);
  Double y=(s-1)/(s+1);
  return((6*zeta[3]+4.*f3m(y)+2.*f3p(y))/ps);
}

// m3s=0, m1s=m2s=m4s=m5s=ms,  ps < 4*ms
static Cplx T12345_special1b(Double ps,Double ms)
{
  Cplx s=csqrt1(1.-4.*ms/ps);
  Cplx y=(s-1)/(s+1);
  return((6*zeta[3]+4.*f3m(y)+2.*f3p(y))/ps);
}

// m3s=0, m1s=m2s=m4s=m5s=ms
static Cplx T12345_special1(Double ps,Double ms)
{
  if (ps<4.*ms) return(T12345_special1b(ps,ms));
  else return(T12345_special1a(ps,ms));
}

// m3s=0, m1s=m4s, m1s=m5s
static Cplx T12345_special2a(Double ps, Double m1s, Double m2s)
{ 
  Double m1=sqrt(m1s), m2=sqrt(m2s);
  Double s=sqrt(1-4.*m1*m2/(ps-sqr(m1-m2)));
  Double y=(s-1)/(s+1);
  Double r=m1/m2;
  return((6*zeta[3]+4.*f3m(y)+4.*f3p(y)
          -f3p(y*r)-f3p(y/r))/ps);
}

// m3s=0, m1s=m4s, m1s=m5s
static Cplx T12345_special2b(Double ps, Double m1s, Double m2s)
{ 
  Double m1=sqrt(m1s), m2=sqrt(m2s);
  Cplx s=csqrt1(1-4.*m1*m2/(ps-sqr(m1-m2)));
  Cplx y=(s-1)/(s+1);
  Double r=m1/m2;
  return((6*zeta[3]+4.*f3m(y)+4.*f3p(y)
          -f3p(y*r)-f3p(y/r))/ps);
}

static Cplx T12345_special2(Double ps, Double m1s, Double m2s)
{
  if (ps>sqr(sqrt(m1s)+sqrt(m2s)))
     return(T12345_special2a(ps,m1s,m2s));
  else
     return(T12345_special2b(ps,m1s,m2s));
}


// m3s=0, m1s=m4s, m1s=m5s=0
static Cplx T12345_special3(Double ps, Double ms)
{
  return((6*zeta[3]-f3p(ms/(ms-ps)))/ps);
}

#ifdef QUADPRECISION
static const doubledouble M1M3M5VALUE("-5.03800310911772511674787191168692108515230");
#else
#define M1M3M5VALUE -5.038003109117725
#endif


Cplx T5aint(Double ps, Double m1s, Double m2s, Double m3s, Double m4s, Double m5s,
            Double prec, Double *abserr)
{
  if ((m3s==Double(0))&&(m1s==m2s)&&(m1s==m4s)&&(m1s==m5s)) {
    *abserr=0; return(T12345_special1(ps,m1s)); };
  if ((m3s==Double(0))&&(m1s==m4s)&&(m2s==m5s)) {
    if ((m1s==Double(0))&&(m2s==Double(0))) {
      Errstream << "special case not implemented: all masses vanishing\n";
      return(Cplx(0)); };
    if (m1s==Double(0)) interchange(&m1s,&m2s);
    if (m2s==Double(0)) {
      *abserr=0; return(T12345_special3(ps,m1s)); };
    *abserr=0; return(T12345_special2(ps,m1s,m2s)); };
  if (  ((ps==m1s)&&(ps==m3s)&&(ps==m5s)&&(m2s==Double(0))&&(m4s==Double(0)))
      ||((ps==m2s)&&(ps==m3s)&&(ps==m4s)&&(m1s==Double(0))&&(m5s==Double(0))) )
    return(M1M3M5VALUE/ps);   // ref. [5]
  if ((m1s==Double(0))||(m2s==Double(0)))
      { interchange(&m1s,&m4s); interchange(&m2s,&m5s); };
  return(T12345integration(ps,m1s,m2s,m3s,m4s,m5s,prec,abserr));
}
