/* ------------------------------------------------------------- */
/* qagutil.h, qagutil.cc                                         */
/* Stefan Bauberger (stefan@bauberger.de)                        */
/* last revision 4.12.95                                         */
/* ------------------------------------------------------------- */
/* Interface for qag.cc,qag.h (adaptive gauss-kronrod-           */
/* integration) to integrate complex functions of real variables */
/* Cplx foo(Double x) in an interval or from a>=0 to infinity    */
/* ------------------------------------------------------------- */

// integration of \int_a^b f(x), with maximal relative error epsrel,
// the estimated achieved absolute error is return in *error:
//
// Cplx Cqag(Cplx (*f) (Double),Double a,Double b,Double epsrel,Double *error);

// integration of \int_a^b f(x) with better convergence in case
// of a sqrt-singularity at f(x=a):
//
// Cplx Cqagasqrtsing(Cplx (*f) (Double),Double a,Double b,
//                    Double epsrel,Double *error);

// integration of \int_a^\infty f(x):
//
// Cplx Cqaginf(Cplx (*f) (Double),Double a,
//              Double epsrel,Double *error); // integr. from a to infinity

// integration of \int_a^\infty f(x) with better convergence in case
// of a sqrt-singularity at f(x=a):
//
// Cplx Cqaginfasqrtsing(Cplx (*f) (Double),Double a,
//                       Double epsrel,Double *error);

// integration of \int_a^\infty f(x)/(x-singe-I*iepssign)
// iepssign = +/-1
//
// Cplx Cqagpinf(Cplx (*f) (Double),Double a,Double singe,int iepssign,
//               Double epsrel,Double *error);

// like Cqagpinf but with better convergence in case
// of a sqrt-singularity at f(x=a):
// Cplx Cqagpinfasqrtsing(Cplx (*f) (Double),Double a,Double singe,int iepssign,
//                        Double epsabs,Double *error);

// integration of f(s)/(s-singe-iepssign*I*eps)^2 to infinity
// Df(x) is supposed to be the derivative of f(x) 
// Cplx Cqagp2inf(Cplx (*f) (Double),Cplx (*Df) (Double),
//                Double a,Double singe,int iepssign,
//                Double epsrel,Double *error);

#include "qagutil.h"
#include "qag.h"

inline Double sqr(const Double& x) { return x*x; }
// inline Cplx sqr(const Cplx& x) { return x*x; }

#define Errstream cerr

/************************************************************/
/* tr2, Dtr2, tr2inv:                                       */
/* transformations which simplify integrations from a       */
/* finite value  y>-.25 to infinity                         */
/************************************************************/
/* a integration of f(x)*Dtr2(x) from tr2inv(s0) to 1.0 is  */
/* equivalent to a integration of f(s) from s0 to infinity  */
/************************************************************/

inline Double tr2(Double x)
{
  return(x/((1-x)*(1-x)));
}

inline Double Dtr2(Double x)   // the derivative of tr2
{
  return((1+x)/((1-x)*(1-x)*(1-x)));
}


inline Double tr2inv(Double y)  // the inverse of tr2
{
  if (y==Double(0)) return(0);
  return(.5*(1.+2.*y-sqrt(1+4.*y))/y);
}

//static CplxInt v;

#ifdef QUADPRECISION
#define MAXQAGITERATION 300
#else
#define MAXQAGITERATION 200
#endif

Cplx Cqag(Cplx (*f) (Double),Double a,Double b,Double epsrel,Double *error)
{
  CplxInt v;
  v.set(0.0,epsrel,MAXQAGITERATION);
  Cplx r=v.integrate(f,a,b);
  *error=v.reported_absolute_error;
  return(r);
}

static Cplx (*fstatictr2)(Double);

static Cplx fstatictr2_call(Double s)
{
  return(fstatictr2(tr2(s))*Dtr2(s));
}

Cplx Cqaginf(Cplx (*f) (Double),Double a,
              Double epsrel,Double *error) // integr. from a to infinity
{
  if (a<=-.25) {
    Errstream << "Cqaginf: Bad integration interval.\n";
    return(Cplx(0)); };
  fstatictr2=f;
  return(Cqag(fstatictr2_call,tr2inv(a),1.,epsrel,error));
}

static Cplx (*fstatictr2b)(Double);

static Cplx fstatictr2b_call(Double s)
{
  return(fstatictr2b(tr2(s))*Dtr2(s));
}

Cplx Cqaginf2(Cplx (*f) (Double),Double a,
              Double epsrel,Double *error) // integr. from a to infinity
{
  if (a<=-.25) {
    Errstream << "Cqaginf: Bad integration interval.\n";
    return(Cplx(0)); };
  fstatictr2b=f;
  return(Cqag(fstatictr2b_call,tr2inv(a),1.,epsrel,error));
}


// calculation of cauchy principal values:

static Cplx (*fstatic1)(Double);
static Double f1_sing,f1_sing2;
static Cplx f1_value0;

static Cplx fstatic1_call(Double s)
{
  return((fstatic1(s)-fstatic1(f1_sing2-s))/(s-f1_sing));
}

static Cplx fstatic1b_call(Double s)
{
  return((fstatic1(s)+fstatic1(f1_sing2-s)-2*f1_value0)/sqr(s-f1_sing));
}

static Cplx fstatic1t_call(Double s)
{
  return(fstatic1(s)/(s-f1_sing));
}

static Cplx fstatic1tb_call(Double s)
{
  return(fstatic1(s)/sqr(s-f1_sing));
}

static Cplx Cqagp1a(Cplx (*f) (Double),Double a,Double singe,int iepssign,
                    Double epsrel,Double *error) 
    // calculation of principal value
    // of f(x)/(x-singe) in the interval from a to b=2*sign-a 
{
  fstatic1=f;
  f1_sing=singe;
  f1_sing2=singe+singe;
  Cplx sc=Cplx(0);
  if (iepssign != 0) sc=iepssign*PiI*f(singe);
  return(Cqag(fstatic1_call,a,singe,epsrel,error)+sc);
}

#define CQPLIMIT 1+1e-15

Cplx Cqagpinf(Cplx (*f) (Double),Double a,Double singe,int iepssign,
               Double epsrel,Double *error)
{
  fstatic1=f;
  f1_sing=singe;
  f1_sing2=singe+singe;
  if (singe<a*CQPLIMIT) return(Cqaginf(fstatic1t_call,a,epsrel,error));
  else {
    f1_sing2=singe+singe;
    Double error1,error2; Cplx r;
    Cplx sc=Cplx(0);
    if (iepssign != 0) sc=iepssign*PiI*f(singe);
    r = Cqag(fstatic1_call,a,singe,epsrel,&error1)
       +Cqaginf(fstatic1t_call,f1_sing2-a,epsrel,&error2)+sc;
    *error=sqrt(sqr(error1)+sqr(error2));
    return(r); };
}

// calculation of cauchy principal value, 2nd set for 2-dim. nested integrals:

static Cplx (*fstatic2)(Double);
static Double f2_sing,f2_sing2;
static Cplx f2_value0;

static Cplx fstatic2_call(Double s)
{
  return((fstatic2(s)-fstatic2(f2_sing2-s))/(s-f2_sing));
}

static Cplx fstatic2t_call(Double s)
{
  return(fstatic2(s)/(s-f2_sing));
}


Cplx Cqagpinf2(Cplx (*f) (Double),Double a,Double singe,int iepssign,
               Double epsrel,Double *error)
{
  fstatic2=f;
  f2_sing=singe;
  f2_sing2=singe+singe;
  if (singe<a*CQPLIMIT) return(Cqaginf2(fstatic2t_call,a,epsrel,error));
  else {
    f2_sing2=singe+singe;
    Double error1,error2; Cplx r;
    Cplx sc=Cplx(0);
    if (iepssign != 0) sc=iepssign*PiI*f(singe);
    r = Cqag(fstatic2_call,a,singe,epsrel,&error1)
       +Cqaginf2(fstatic2t_call,f2_sing2-a,epsrel,&error2)+sc;
    *error=sqrt(sqr(error1)+sqr(error2));
    return(r); };
}


static Cplx (*fstatic3)(Double);
static Double f3_sing;

static Cplx fstatic3_call(Double s)
{
  return(fstatic3(s)/(s-f3_sing));
}

Cplx Cqagp1(Cplx (*f) (Double),Double a,Double b,Double singe,int iepssign,
                   Double epsrel,Double *error) 
    // calculation of principal value
    // of f(x)/(x-singe) in the interval from a to b 
{
  fstatic3=f;
  f3_sing=singe;
  if ((singe<a*CQPLIMIT)||(singe>b/CQPLIMIT))
    return(Cqag(fstatic3_call,a,b,epsrel,error));
  Double error1,error2;
  Cplx res;
  if ((b+a)<2*singe) {
    Double l=2*singe-b;
    res=Cqag(fstatic3_call,a,l,epsrel,&error1);
    res+=Cqagp1a(f,l,singe,iepssign,epsrel,&error2);
  }
  else {
    Double l=2*singe-a;
    res=Cqag(fstatic3_call,l,b,epsrel,&error1);
    res+=Cqagp1a(f,a,singe,iepssign,epsrel,&error2);
  };
  *error=sqrt(sqr(error1)+sqr(error2));
  return(res);
}

Cplx Cqagpinf(Cplx (*f) (Double),Double a,Double singe,int iepssign,
              Double epsrel,Double *error,Double limit1,Double limit2)
     // like previous Cqagpinf but take limit1 and limit2 as borders of
     // integration intervals
{
  if (limit1>limit2) {
    Double x=limit1; limit1=limit2; limit2=x; };
  if (a>=limit2) return(Cqagpinf(f,a,singe,iepssign,epsrel,error));
  Double error1,error2=0,error3=0;
  Cplx res=Cqagpinf(f,limit2,singe,iepssign,epsrel,&error1);
  if (a>=limit1) res+=Cqagp1(f,a,limit2,singe,iepssign,epsrel,&error2);
  else {
    res+=Cqagp1(f,limit1,limit2,singe,iepssign,epsrel,&error2);
    res+=Cqagp1(f,a,limit1,singe,iepssign,epsrel,&error3); };
  *error=sqrt(sqr(error1)+sqr(error2)+sqr(error3));
  return(res);
}


static Cplx (*fasqrtsing)(Double);

static Double a0sqrtsing;

static Cplx fasqrtsing_call(Double t)
{
  Double s=sqr(t)+a0sqrtsing;
  return(2*t*fasqrtsing(s));
}

// for the case of a sqrt singularity of f at a:
Cplx Cqagasqrtsing(Cplx (*f) (Double),Double a,Double b,
                      Double epsrel,Double *error)
{
  a0sqrtsing=a;
  fasqrtsing=f;
  return(Cqag(fasqrtsing_call,0,sqrt(b-a),epsrel,error));
}

Cplx Cqaginfasqrtsing(Cplx (*f) (Double),Double a,
                      Double epsrel,Double *error)
{
  a0sqrtsing=a;
  fasqrtsing=f;
  return(Cqaginf(fasqrtsing_call,0,epsrel,error));
}


static Cplx (*fbsqrtsing)(Double);

static Double b0sqrtsing;

static Cplx fbsqrtsing_call(Double t)
{
  Double s=sqr(t)+b0sqrtsing;
  return(2*t*fbsqrtsing(s));
}

// for the case of a sqrt singularity of f at a:
Cplx Cqagasqrtsing2(Cplx (*f) (Double),Double a,Double b,
                      Double epsrel,Double *error)
{
  b0sqrtsing=a;
  fbsqrtsing=f;
  return(Cqag(fbsqrtsing_call,0,sqrt(b-a),epsrel,error));
}

Cplx Cqaginfasqrtsing2(Cplx (*f) (Double),Double a,
                       Double epsrel,Double *error)
{
  b0sqrtsing=a;
  fbsqrtsing=f;
  return(Cqaginf2(fbsqrtsing_call,0,epsrel,error));
}


static Cplx fsqrtsingstatic1_call(Double t)
{
  Double s=sqr(t)+a0sqrtsing;
  return(2*(t*fstatic1(s)-t*fstatic1(f1_sing2-s))
               /(s-f1_sing));
}

static Cplx fsqrtsingstatic2_call(Double t)
{
  Double s=sqr(t)+a0sqrtsing;
  return(2*t*fstatic1(s)/(s-f1_sing));
}

Cplx Cqagpinfasqrtsing(Cplx (*f) (Double),Double a,Double singe,int iepssign,
                        Double epsrel,Double *error)
{
  a0sqrtsing=a;
  fstatic1=f;
  f1_sing= singe;
  f1_sing2=2*singe;
  if (singe < a*CQPLIMIT)
    return(Cqaginf(fsqrtsingstatic2_call,0,epsrel,error));
  else {
    Double error1,error2; Cplx r;
    Cplx sc=Cplx(0);
    if (iepssign != 0) sc=iepssign*PiI*f(singe);
    r = Cqag(fsqrtsingstatic1_call,0,sqrt(singe-a),epsrel,&error1)
       +Cqaginf(fstatic1t_call,f1_sing2-a,epsrel,&error2)+sc;
    *error=sqrt(sqr(error1)+sqr(error2));
    return(r); };
}


Cplx Cqagp2inf(Cplx (*f) (Double),Cplx (*Df) (Double),
		Double a,Double singe,int iepssign,
                Double epsrel,Double *error)
       // integration of f(s)/(s-singe-iepssign*I*eps)^2 to infinity
       // Df(x) is supposed to be the derivative of f(x) 
{
  fstatic1=f;
  f1_sing=singe;
  if (singe<=a*CQPLIMIT) return(Cqaginf(fstatic1tb_call,a,epsrel,error));
  f1_sing2=singe+singe;
  f1_value0=(*f)(singe);
  Double error1,error2;
  Cplx r1,r2,sc=Cplx(0);
  r1 = Cqag(fstatic1b_call,a,singe,epsrel,&error1);
  r2 = Cqaginf(fstatic1tb_call,f1_sing2-a,epsrel,&error2);
  if (iepssign != 0) sc=iepssign*PiI*Df(singe);
  *error=sqrt(sqr(error1)+sqr(error2));
  return(r1+r2+sc+2*f1_value0/(a-singe));
}

static Cplx Cqagp2a(Cplx (*f) (Double),Cplx (*Df) (Double),
                    Double a,Double singe,int iepssign,
                    Double epsrel,Double *error) 
    // calculation of principal value
    // of f(x)/(x-singe)^2 in the interval from a to b=2*singe-a
{
  fstatic1=f;
  f1_sing=singe;
  f1_sing2=singe+singe;
  f1_value0=(*f)(singe);
  Cplx r1;
  r1 = Cqag(fstatic1b_call,a,singe,epsrel,error);
  if (iepssign != 0) r1+=iepssign*PiI*Df(singe);
  return(r1+4*f1_value0/(a-singe));
}


static Cplx (*fstatic4)(Double);
static Double f4_sing;

static Cplx fstatic4_call(Double s)
{
  return(fstatic4(s)/sqr(s-f4_sing));
}

Cplx Cqagp2(Cplx (*f) (Double),Cplx (*Df) (Double),
            Double a,Double b,Double singe,int iepssign,
            Double epsrel,Double *error) 
    // calculation of principal value
    // of f(x)/(x-singe)^2 in the interval from a to b 
{
  fstatic4=f;
  f4_sing=singe;
  if ((singe<a*CQPLIMIT)||(singe>b/CQPLIMIT))
    return(Cqag(fstatic4_call,a,b,epsrel,error));
  Double error1,error2;
  Cplx res;
  if ((b+a)<2*singe) {
    Double l=2*singe-b;
    res=Cqag(fstatic4_call,a,l,epsrel,&error1);
    res+=Cqagp2a(f,Df,l,singe,iepssign,epsrel,&error2);
  }
  else {
    Double l=2*singe-a;
    res=Cqag(fstatic4_call,l,b,epsrel,&error1);
    res+=Cqagp2a(f,Df,a,singe,iepssign,epsrel,&error2);
  };
  *error=sqrt(sqr(error1)+sqr(error2));
  return(res);
}


Cplx Cqagp2inf(Cplx (*f) (Double),Cplx (*Df) (Double),
               Double a,Double singe,int iepssign,
               Double epsrel,Double *error,Double limit1,Double limit2)
     // like previous Cqagp2inf but take limit1 and limit2 as borders of
     // integration intervals
{
  if (limit1>limit2) {
    Double x=limit1; limit1=limit2; limit2=x; };
  if (a>=limit2) return(Cqagp2inf(f,Df,a,singe,iepssign,epsrel,error));
  Double error1,error2=0,error3=0;
  Cplx res=Cqagp2inf(f,Df,limit2,singe,iepssign,epsrel,&error1);
  cerr << " - 2 - \n";
  if (a>=limit1) res+=Cqagp2(f,Df,a,limit2,singe,iepssign,epsrel,&error2);
  else {
    res+=Cqagp2(f,Df,limit1,limit2,singe,iepssign,epsrel,&error2);
    res+=Cqagp2(f,Df,a,limit1,singe,iepssign,epsrel,&error3); };
  *error=sqrt(sqr(error1)+sqr(error2)+sqr(error3));
  return(res);
}

