/* ------------------------------------------------------------- */
/* TVID: Three-loop Vacuum Integrals from Dispersion relations   */
/* version 2.0 (Aug 23, 2019)                                     */
/* ------------------------------------------------------------- */
/* ucall.cc                                                      */
/* Ayres Freitas (afreitas@pitt.edu)			         */
/* last revision 23.08.19                                        */
/* ------------------------------------------------------------- */
/* Main program for calling C(++) routines for evaluation of     */
/* finite parts of Uxx and Txx integrals               	         */
/* ------------------------------------------------------------- */
/* References:                                                   */
/*   [1] = Freitas, arXiv:1609.09159, JHEP 1611 (2016) 145       */
/*   [2] = Bauberger, Freitas, Wiegand, arXiv:1908.09887         */
/* ------------------------------------------------------------- */

#include <iostream>
#include <fstream>
#include <stdio.h>
#include <string.h>
using namespace std;
	
#include "Cplx.h"
#include "B0.h"
#include "U456.h"
#include "Uabc.h"
#include "Umn.h"
#include "U78.h"
#include "T345.h"
#include "clooptools.h"


#ifdef QUADPRECISION
Double Cprec=1e-12;
#else
Double Cprec=1e-6;
#endif

#ifdef QUADPRECISION
static doubledouble EPS=SDouble("1e-2");
#else
#define EPS 1e-2
#endif


int main(int argc, char *argv[])
{
  int prerr = 0;
  
  cout.precision(16);

  if(argc < 3)
  {
    cerr << "Error: no input/output file specified" << endl;
    return 1;
  }
  if(argc > 3 && strcmp(argv[3], "-e")==0)
    prerr = 1;
  
  char line[500], *strp;
  Double pars[9];
  unsigned pc;
  Double abserr, err1, err2, err3, err4;
  Cplx ires;
  
  ifstream inf(argv[1], ios::in );
  if(!inf)
  {
    cerr << "Error in input file " << argv[1] << endl;
    exit(1);
  }
  ofstream ouf(argv[2], ios::out );
  if(!ouf)
  {
    cerr << "Error in output file " << argv[2] << endl;
    exit(1);
  }
  ouf.precision(16);
//  ouf.precision(30);

  while(inf.getline(line, sizeof(line)-1))
  {
    strp = strtok(line, " \t");
    cout << strp << endl;
    strp = strtok(NULL, " \t");
    for(pc = 0; pc < sizeof(line) && strp != NULL; pc++)
    {
      pars[pc] = atodd(strp);
      strp = strtok(NULL, " \t");
    }
    if(strp != NULL)
      cerr << "Too many parameters in\n " << line << endl;
    strp = strtok(line, " \t");

    if(strcmp(strp, "B0fin") == 0 && pc == 3)
    {
      ires = B0A(pars[0], pars[1], pars[2]) +2;
      abserr = 0;
    }
    if(strcmp(strp, "T3a") == 0 && pc == 4)
      ires = T3aint(pars[0], pars[1], pars[2], pars[3], Cprec, &abserr);
    if(strcmp(strp, "T3a1") == 0 && pc == 4)
      ires = T3a1int(pars[0], pars[1], pars[2], pars[3], Cprec, &abserr);
    if(strcmp(strp, "T4a") == 0 && pc == 5)
      ires = T4aint(pars[0], pars[1], pars[2], pars[3], pars[4], Cprec, &abserr);
    if(strcmp(strp, "T4a1") == 0 && pc == 5)
      ires = T4a1int(pars[0], pars[1], pars[2], pars[3], pars[4], Cprec, &abserr);
    if(strcmp(strp, "T5a") == 0 && pc == 6)
      ires = T5aint(pars[0], pars[1], pars[2], pars[3], pars[4], pars[5], 
      		   Cprec, &abserr);
    if(strcmp(strp, "U4") == 0 && pc == 4)
      ires = U4int(pars[0], pars[1], pars[2], pars[3], Cprec, &abserr);
    if(strcmp(strp, "U40") == 0 && pc == 3)
      ires = U40int(pars[0], pars[1], pars[2], Cprec*10, &abserr);
    if(strcmp(strp, "U5") == 0 && pc == 5)
      ires = U5int(pars[0], pars[1], pars[2], pars[3], pars[4], Cprec, &abserr);
    if(strcmp(strp, "U50") == 0 && pc == 3)
      ires = U50int(pars[0], pars[1], pars[2], Cprec, &abserr);
    if(strcmp(strp, "U6") == 0 && pc == 6)
      ires = U6int(pars[0], pars[1], pars[2], pars[3], pars[4], pars[5],
                   Cprec, &abserr);
    if(strcmp(strp, "U4a") == 0 && pc == 5)
      ires = U4aint(pars[0], pars[1], pars[2], pars[3], pars[4], Cprec, &abserr);
    if(strcmp(strp, "U5a") == 0 && pc == 6)
      ires = U5aint(pars[0], pars[1], pars[2], pars[3], pars[4], pars[5],
                    Cprec, &abserr);
    if(strcmp(strp, "U5b") == 0 && pc == 6)
      ires = U5bint(pars[0], pars[1], pars[2], pars[3], pars[4], pars[5],
                    Cprec, &abserr);
    if(strcmp(strp, "U5c") == 0 && pc == 6)
      ires = U5cint(pars[0], pars[1], pars[2], pars[3], pars[4], pars[5],
                    Cprec, &abserr);
    if(strcmp(strp, "U6a") == 0 && pc == 7)
      ires = U6aint(pars[0], pars[1], pars[2], pars[3], pars[4], pars[5], pars[6],
                    Cprec, &abserr);
    if(strcmp(strp, "U6b") == 0 && pc == 7)
      ires = U6bint(pars[0], pars[1], pars[2], pars[3], pars[4], pars[5], pars[6],
                    Cprec, &abserr);
    if(strcmp(strp, "U6c") == 0 && pc == 7)
      ires = U6cint(pars[0], pars[1], pars[2], pars[3], pars[4], pars[5], pars[6],
                    Cprec, &abserr);
    if(strcmp(strp, "U6m") == 0 && pc == 7)
      ires = U6mint(pars[0], pars[1], pars[2], pars[3], pars[4], pars[5], pars[6],
                    Cprec, &abserr);
    if(strcmp(strp, "U6n") == 0 && pc == 7)
      ires = U6nint(pars[0], pars[1], pars[2], pars[3], pars[4], pars[5], pars[6],
                    Cprec, &abserr);
    if(strcmp(strp, "U7m") == 0 && pc == 8)
      ires = U7mint(pars[0], pars[1], pars[2], pars[3], pars[4], pars[5], pars[6],
                    pars[7], Cprec, &abserr);
    if(strcmp(strp, "U4a1") == 0 && pc == 5)
      ires = U4a1int(pars[0], pars[1], pars[2], pars[3], pars[4], Cprec, &abserr);
    if(strcmp(strp, "U4a2") == 0 && pc == 5)
      ires = U4a2int(pars[0], pars[1], pars[2], pars[3], pars[4], Cprec, &abserr);
    if(strcmp(strp, "U4a3") == 0 && pc == 5)
      ires = U4a3int(pars[0], pars[1], pars[2], pars[3], pars[4], Cprec, &abserr);
    if(strcmp(strp, "U5a1") == 0 && pc == 6)
      ires = U5a1int(pars[0], pars[1], pars[2], pars[3], pars[4], pars[5],
                     Cprec, &abserr);
    if(strcmp(strp, "U5a2") == 0 && pc == 6)
      ires = U5a2int(pars[0], pars[1], pars[2], pars[3], pars[4], pars[5],
                     Cprec, &abserr);
    if(strcmp(strp, "U5b1") == 0 && pc == 6)
      ires = U5b1int(pars[0], pars[1], pars[2], pars[3], pars[4], pars[5],
                     Cprec, &abserr);
    if(strcmp(strp, "U5b2") == 0 && pc == 6)
      ires = U5b2int(pars[0], pars[1], pars[2], pars[3], pars[4], pars[5],
                     Cprec, &abserr);
    if(strcmp(strp, "U5c1") == 0 && pc == 6)
      ires = U5c1int(pars[0], pars[1], pars[2], pars[3], pars[4], pars[5],
                     Cprec, &abserr);
    if(strcmp(strp, "U6m2") == 0 && pc == 7)
      ires = U6m2int(pars[0], pars[1], pars[2], pars[3], pars[4], pars[5], pars[6],
                     Cprec, &abserr);
    if(strcmp(strp, "U6n1") == 0 && pc == 7)
      ires = U6n1int(pars[0], pars[1], pars[2], pars[3], pars[4], pars[5], pars[6],
                     Cprec, &abserr);

    if(strcmp(strp, "U6m1") == 0 && pc == 7)
    {
      ires = (-U6mint(pars[0], pars[1], pars[2], pars[3]*(1+2*EPS), 
      		        pars[4], pars[5], pars[6], Cprec, &err1)
	      +8*U6mint(pars[0], pars[1], pars[2], pars[3]*(1+EPS), 
      		        pars[4], pars[5], pars[6], Cprec, &err2)
	      -8*U6mint(pars[0], pars[1], pars[2], pars[3]*(1-EPS), 
      		        pars[4], pars[5], pars[6], Cprec, &err3)
	      +U6mint(pars[0], pars[1], pars[2], pars[3]*(1-2*EPS), 
      		        pars[4], pars[5], pars[6], Cprec, &err4))/(12*pars[3]*EPS);
      abserr = sqrt(sqr(err1) + sqr(err2*8) + sqr(err3*8) + sqr(err4)) /
                (12*pars[3]*EPS);
    }
    if(strcmp(strp, "U6m3") == 0 && pc == 7)
    {
      ires = (-U6mint(pars[0], pars[1], pars[2]*(1+2*EPS), pars[3], 
      		        pars[4], pars[5], pars[6], Cprec, &err1)
	      +8*U6mint(pars[0], pars[1], pars[2]*(1+EPS), pars[3], 
      		        pars[4], pars[5], pars[6], Cprec, &err2)
	      -8*U6mint(pars[0], pars[1], pars[2]*(1-EPS), pars[3], 
      		        pars[4], pars[5], pars[6], Cprec, &err3)
	      +U6mint(pars[0], pars[1], pars[2]*(1-2*EPS), pars[3], 
      		        pars[4], pars[5], pars[6], Cprec, &err4))/(12*pars[2]*EPS);
      abserr = sqrt(sqr(err1) + sqr(err2*8) + sqr(err3*8) + sqr(err4)) /
                (12*pars[2]*EPS);
    }
    if(strcmp(strp, "U6n2") == 0 && pc == 7)
    {
      ires = (-U6nint(pars[0], pars[1], pars[2]*(1+2*EPS), pars[3], 
      		        pars[4], pars[5], pars[6], Cprec, &err1)
	      +8*U6nint(pars[0], pars[1], pars[2]*(1+EPS), pars[3], 
      		        pars[4], pars[5], pars[6], Cprec, &err2)
	      -8*U6nint(pars[0], pars[1], pars[2]*(1-EPS), pars[3], 
      		        pars[4], pars[5], pars[6], Cprec, &err3)
	      +U6nint(pars[0], pars[1], pars[2]*(1-2*EPS), pars[3], 
      		        pars[4], pars[5], pars[6], Cprec, &err4))/(12*pars[2]*EPS);
      abserr = sqrt(sqr(err1) + sqr(err2*8) + sqr(err3*8) + sqr(err4)) /
                (12*pars[2]*EPS);
    }
    if(strcmp(strp, "U7a") == 0 && pc == 8)
    {
      ltini();
      Cprec = 1e-10;
      ires = U7aint(pars[0], pars[1], pars[2], pars[3], pars[4], pars[5],
                    pars[6], pars[7], Cprec, &abserr);
    }
/*    if(strcmp(strp, "U7a1") == 0 && pc == 8)
    {
      ltini();
      Cprec = 1e-8;
      ires = U7a1int(pars[0], pars[1], pars[2], pars[3], pars[4], pars[5],
                     pars[6], pars[7], Cprec, &abserr);
    }
    if(strcmp(strp, "U7a2") == 0 && pc == 8)
    {
      ltini();
      Cprec = 1e-8;
      ires = U7a2int(pars[0], pars[1], pars[2], pars[3], pars[4], pars[5],
                     pars[6], pars[7], Cprec, &abserr);
    }
*/
    if(strcmp(strp, "U7a1") == 0 && pc == 8)
    {
      ltini();
      Cprec = 1e-10;
      ires = (-U7aint(pars[0], pars[1], pars[2], pars[3]*(1+2*EPS), 
      		        pars[4], pars[5], pars[6], pars[7], Cprec, &err1)
	      +8*U7aint(pars[0], pars[1], pars[2], pars[3]*(1+EPS), 
      		        pars[4], pars[5], pars[6], pars[7], Cprec, &err2)
	      -8*U7aint(pars[0], pars[1], pars[2], pars[3]*(1-EPS), 
      		        pars[4], pars[5], pars[6], pars[7], Cprec, &err3)
	      +U7aint(pars[0], pars[1], pars[2], pars[3]*(1-2*EPS), 
      		        pars[4], pars[5], pars[6], pars[7], Cprec, &err4))/(12*pars[3]*EPS);
      abserr = sqrt(sqr(err1) + sqr(err2*8) + sqr(err3*8) + sqr(err4)) /
                (12*pars[2]*EPS);
    }
    if(strcmp(strp, "U7a2") == 0 && pc == 8)
    {
      ltini();
      Cprec = 1e-10;
      ires = (-U7aint(pars[0], pars[1], pars[2]*(1+2*EPS), pars[3], 
      		        pars[4], pars[5], pars[6], pars[7], Cprec, &err1)
	      +8*U7aint(pars[0], pars[1], pars[2]*(1+EPS), pars[3], 
      		        pars[4], pars[5], pars[6], pars[7], Cprec, &err2)
	      -8*U7aint(pars[0], pars[1], pars[2]*(1-EPS), pars[3], 
      		        pars[4], pars[5], pars[6], pars[7], Cprec, &err3)
	      +U7aint(pars[0], pars[1], pars[2]*(1-2*EPS), pars[3], 
      		        pars[4], pars[5], pars[6], pars[7], Cprec, &err4))/(12*pars[2]*EPS);
      abserr = sqrt(sqr(err1) + sqr(err2*8) + sqr(err3*8) + sqr(err4)) /
                (12*pars[2]*EPS);
    }
    if(strcmp(strp, "U8a") == 0 && pc == 9)
    {
      ltini();
      Cprec = 1e-8;
      ires = U8aint(pars[0], pars[1], pars[2], pars[3], pars[4], pars[5],
                   pars[6], pars[7], pars[8], Cprec, &abserr);
    }
    
    if(prerr)
      ouf << ires.real() << " " << ires.imag() << " " << abserr << endl;
    else
      ouf << ires.real() << " " << ires.imag() << endl;
  }
  inf.close();
  ouf.close();

  return 0;
}
